'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var gLite = require('@antv/g-lite');
var util = require('@antv/util');

function _extends() {
  _extends = Object.assign ? Object.assign.bind() : function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];
      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }
    return target;
  };
  return _extends.apply(this, arguments);
}
function _inheritsLoose(subClass, superClass) {
  subClass.prototype = Object.create(superClass.prototype);
  subClass.prototype.constructor = subClass;
  _setPrototypeOf(subClass, superClass);
}
function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf ? Object.setPrototypeOf.bind() : function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };
  return _setPrototypeOf(o, p);
}
function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;
  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }
  return target;
}

/**
 * Common utilities
 * @module glMatrix
 */
var ARRAY_TYPE = typeof Float32Array !== 'undefined' ? Float32Array : Array;
if (!Math.hypot) Math.hypot = function () {
  var y = 0,
      i = arguments.length;

  while (i--) {
    y += arguments[i] * arguments[i];
  }

  return Math.sqrt(y);
};

/**
 * 3 Dimensional Vector
 * @module vec3
 */

/**
 * Creates a new, empty vec3
 *
 * @returns {vec3} a new 3D vector
 */

function create() {
  var out = new ARRAY_TYPE(3);

  if (ARRAY_TYPE != Float32Array) {
    out[0] = 0;
    out[1] = 0;
    out[2] = 0;
  }

  return out;
}
/**
 * Creates a new vec3 initialized with the given values
 *
 * @param {Number} x X component
 * @param {Number} y Y component
 * @param {Number} z Z component
 * @returns {vec3} a new 3D vector
 */

function fromValues(x, y, z) {
  var out = new ARRAY_TYPE(3);
  out[0] = x;
  out[1] = y;
  out[2] = z;
  return out;
}
/**
 * Subtracts vector b from vector a
 *
 * @param {vec3} out the receiving vector
 * @param {ReadonlyVec3} a the first operand
 * @param {ReadonlyVec3} b the second operand
 * @returns {vec3} out
 */

function subtract(out, a, b) {
  out[0] = a[0] - b[0];
  out[1] = a[1] - b[1];
  out[2] = a[2] - b[2];
  return out;
}
/**
 * Alias for {@link vec3.subtract}
 * @function
 */

var sub = subtract;
/**
 * Perform some operation over an array of vec3s.
 *
 * @param {Array} a the array of vectors to iterate over
 * @param {Number} stride Number of elements between the start of each vec3. If 0 assumes tightly packed
 * @param {Number} offset Number of elements to skip at the beginning of the array
 * @param {Number} count Number of vec3s to iterate over. If 0 iterates over entire array
 * @param {Function} fn Function to call for each vector in the array
 * @param {Object} [arg] additional argument to pass to fn
 * @returns {Array} a
 * @function
 */

var forEach = function () {
  var vec = create();
  return function (a, stride, offset, count, fn, arg) {
    var i, l;

    if (!stride) {
      stride = 3;
    }

    if (!offset) {
      offset = 0;
    }

    if (count) {
      l = Math.min(count * stride + offset, a.length);
    } else {
      l = a.length;
    }

    for (i = offset; i < l; i += stride) {
      vec[0] = a[i];
      vec[1] = a[i + 1];
      vec[2] = a[i + 2];
      fn(vec, vec, arg);
      a[i] = vec[0];
      a[i + 1] = vec[1];
      a[i + 2] = vec[2];
    }

    return a;
  };
}();

var _excluded = ["body", "startHead", "endHead", "startHeadOffset", "endHeadOffset"],
  _excluded2 = ["body", "startHead", "endHead", "startHeadOffset", "endHeadOffset"],
  _excluded3 = ["body", "startHead", "endHead", "startHeadOffset", "endHeadOffset"];
/**
 * support 3 types of arrow line:
 * 1. Line
 * 2. Polyline
 * 3. Path
 *
 * support 2 types of arrow head:
 * 1. default(Path)
 * 2. custom
 */
var Arrow = /*#__PURE__*/function (_CustomElement) {
  _inheritsLoose(Arrow, _CustomElement);
  function Arrow(config) {
    var _this;
    _this = _CustomElement.call(this, _extends({}, config, {
      type: Arrow.tag
    })) || this;
    _this.body = void 0;
    _this.startHead = void 0;
    _this.endHead = void 0;
    _this.startHeadPosition = void 0;
    _this.startHeadRad = void 0;
    _this.endHeadPosition = void 0;
    _this.endHeadRad = void 0;
    var _this$attributes = _this.attributes,
      body = _this$attributes.body,
      startHead = _this$attributes.startHead,
      endHead = _this$attributes.endHead,
      startHeadOffset = _this$attributes.startHeadOffset,
      endHeadOffset = _this$attributes.endHeadOffset,
      rest = _objectWithoutPropertiesLoose(_this$attributes, _excluded);
    if (!body) {
      throw new Error("Arrow's body is required");
    }
    // append arrow body
    _this.body = body;
    _this.appendChild(_this.body);
    _this.handleBodyAttributeChanged(_this.body);
    if (startHead) {
      _this.appendArrowHead(_this.getArrowHeadType(startHead), true);
    }
    if (endHead) {
      _this.appendArrowHead(_this.getArrowHeadType(endHead), false);
    }
    _this.applyArrowStyle(rest, [_this.body, _this.startHead, _this.endHead]);
    return _this;
  }
  var _proto = Arrow.prototype;
  _proto.handleBodyAttributeChanged = function handleBodyAttributeChanged(body) {
    var _this2 = this;
    body.addEventListener(gLite.ElementEvent.ATTR_MODIFIED, function (e) {
      var attrName = e.attrName;
      if (attrName === 'x1' || attrName === 'y1') {
        if (_this2.startHead) {
          _this2.transformArrowHead(_this2.startHead, true);
        }
      } else if (attrName === 'x2' || attrName === 'y2') {
        if (_this2.endHead) {
          _this2.transformArrowHead(_this2.endHead, false);
        }
      }
      // const { nodeName } = body;
      // if (nodeName === Shape.LINE) {
      //   body
      // }
    });
  };
  _proto.getBody = function getBody() {
    return this.body;
  };
  _proto.getStartHead = function getStartHead() {
    return this.startHead;
  };
  _proto.getEndHead = function getEndHead() {
    return this.endHead;
  };
  _proto.attributeChangedCallback = function attributeChangedCallback(name, oldValue, newValue) {
    if (name === 'opacity' || name === 'strokeOpacity' || name === 'stroke' || name === 'lineWidth' || name === 'increasedLineWidthForHitTesting') {
      var _this$applyArrowStyle;
      this.applyArrowStyle((_this$applyArrowStyle = {}, _this$applyArrowStyle[name] = newValue, _this$applyArrowStyle), [this.body, this.startHead, this.endHead]);
    } else if (name === 'startHead' || name === 'endHead') {
      var isStart = name === 'startHead';
      // delete existed arrow head first
      this.destroyArrowHead(isStart);
      if (newValue) {
        var _this$attributes2 = this.attributes,
          body = _this$attributes2.body,
          startHead = _this$attributes2.startHead,
          endHead = _this$attributes2.endHead,
          startHeadOffset = _this$attributes2.startHeadOffset,
          endHeadOffset = _this$attributes2.endHeadOffset,
          rest = _objectWithoutPropertiesLoose(_this$attributes2, _excluded2);
        // append new arrow head
        this.appendArrowHead(this.getArrowHeadType(newValue), isStart);
        this.applyArrowStyle(rest, [isStart ? this.startHead : this.endHead]);
      }
    } else if (name === 'body') {
      var _this$attributes3 = this.attributes,
        _body = _this$attributes3.body,
        _startHead = _this$attributes3.startHead,
        _endHead = _this$attributes3.endHead,
        _startHeadOffset = _this$attributes3.startHeadOffset,
        _endHeadOffset = _this$attributes3.endHeadOffset,
        _rest = _objectWithoutPropertiesLoose(_this$attributes3, _excluded3);
      this.body.destroy();
      // @ts-ignore
      this.body = newValue;
      this.appendChild(this.body);
      this.applyArrowStyle(_rest, [this.body]);
    } else if (name === 'startHeadOffset') {
      this.moveArrowHeadAlongTangent(newValue, true);
    } else if (name === 'endHeadOffset') {
      this.moveArrowHeadAlongTangent(newValue, false);
    }
  };
  _proto.getArrowHeadType = function getArrowHeadType(head) {
    if (typeof head === 'boolean') {
      return 'default';
    }
    return 'custom';
  };
  _proto.appendArrowHead = function appendArrowHead(type, isStart) {
    var head;
    if (type === 'default') {
      head = this.createDefaultArrowHead();
    } else {
      // @ts-ignore
      head = isStart ? this.attributes.startHead : this.attributes.endHead;
    }
    // set position & rotation
    this.transformArrowHead(head, isStart);
    // heads should display on top of body
    head.setAttribute('zIndex', 1);
    if (isStart) {
      this.startHead = head;
    } else {
      this.endHead = head;
    }
    this.appendChild(head);
    var offset = isStart ? this.attributes.startHeadOffset : this.attributes.endHeadOffset;
    if (offset) {
      this.moveArrowHeadAlongTangent(offset, isStart);
    }
  }
  /**
   * transform arrow head according to arrow line
   */;
  _proto.transformArrowHead = function transformArrowHead(head, isStart) {
    var position = create();
    var rad = 0;
    var x1 = 0;
    var x2 = 0;
    var y1 = 0;
    var y2 = 0;
    var bodyType = this.body && this.body.nodeName;
    if (bodyType === gLite.Shape.LINE) {
      var _this$body$attributes = this.body.attributes,
        _x1 = _this$body$attributes.x1,
        _x2 = _this$body$attributes.x2,
        _y1 = _this$body$attributes.y1,
        _y2 = _this$body$attributes.y2;
      x1 = isStart ? _x2 : _x1;
      x2 = isStart ? _x1 : _x2;
      y1 = isStart ? _y2 : _y1;
      y2 = isStart ? _y1 : _y2;
    } else if (bodyType === gLite.Shape.POLYLINE) {
      var points = this.body.attributes.points;
      var length = points.length;
      x1 = isStart ? points[1][0] : points[length - 2][0];
      y1 = isStart ? points[1][1] : points[length - 2][1];
      x2 = isStart ? points[0][0] : points[length - 1][0];
      y2 = isStart ? points[0][1] : points[length - 1][1];
    } else if (bodyType === gLite.Shape.PATH) {
      var _this$getTangent = this.getTangent(this.body, isStart),
        p1 = _this$getTangent[0],
        p2 = _this$getTangent[1];
      x1 = p1[0];
      y1 = p1[1];
      x2 = p2[0];
      y2 = p2[1];
    }
    var x = x1 - x2;
    var y = y1 - y2;
    rad = Math.atan2(y, x);
    position = fromValues(x2, y2, 0);
    if (isStart) {
      this.startHeadPosition = position;
      this.startHeadRad = rad;
    } else {
      this.endHeadPosition = position;
      this.endHeadRad = rad;
    }
    head.setLocalPosition(position);
    head.setLocalEulerAngles(rad * 180 / Math.PI + head.getLocalEulerAngles());
  };
  _proto.moveArrowHeadAlongTangent = function moveArrowHeadAlongTangent(offset, isStart) {
    var head = isStart ? this.startHead : this.endHead;
    if (head) {
      head.setLocalPosition(sub(create(), isStart ? this.startHeadPosition : this.endHeadPosition, fromValues(Math.cos(isStart ? this.startHeadRad : this.endHeadRad) * offset, Math.sin(isStart ? this.startHeadRad : this.endHeadRad) * offset, 0)));
    }
    // cut body
    if (this.body) ;
  };
  _proto.destroyArrowHead = function destroyArrowHead(isStart) {
    if (isStart && this.startHead) {
      this.startHead.destroy();
      this.startHead = undefined;
    }
    if (!isStart && this.endHead) {
      this.endHead.destroy();
      this.endHead = undefined;
    }
  };
  _proto.getTangent = function getTangent(path, isStart) {
    return isStart ? path.getStartTangent() : path.getEndTangent();
  };
  _proto.createDefaultArrowHead = function createDefaultArrowHead() {
    var _this$attributes4 = this.attributes,
      stroke = _this$attributes4.stroke,
      lineWidth = _this$attributes4.lineWidth;
    var sin = Math.sin,
      cos = Math.cos,
      PI = Math.PI;
    return new gLite.Path({
      style: {
        // draw an angle '<'
        path: "M" + 10 * cos(PI / 6) + "," + 10 * sin(PI / 6) + " L0,0 L" + 10 * cos(PI / 6) + ",-" + 10 * sin(PI / 6),
        stroke: stroke,
        lineWidth: lineWidth,
        transformOrigin: 'center',
        anchor: [0.5, 0.5] // set anchor to center
      }
    });
  };
  _proto.applyArrowStyle = function applyArrowStyle(attributes, objects) {
    var opacity = attributes.opacity,
      stroke = attributes.stroke,
      strokeOpacity = attributes.strokeOpacity,
      lineWidth = attributes.lineWidth,
      increasedLineWidthForHitTesting = attributes.increasedLineWidthForHitTesting;
    objects.forEach(function (shape) {
      if (shape) {
        if (!util.isNil(opacity)) {
          shape.style.opacity = opacity;
        }
        if (!util.isNil(stroke)) {
          shape.style.stroke = stroke;
        }
        if (!util.isNil(strokeOpacity)) {
          shape.style.strokeOpacity = strokeOpacity;
        }
        if (!util.isNil(lineWidth)) {
          shape.style.lineWidth = lineWidth;
        }
        if (!util.isNil(increasedLineWidthForHitTesting)) {
          shape.style.increasedLineWidthForHitTesting = increasedLineWidthForHitTesting;
        }
      }
    });
  };
  return Arrow;
}(gLite.CustomElement);
Arrow.tag = 'arrow';

var PI = Math.PI;
var PI2 = PI * 2;
var mathSin = Math.sin;
var mathCos = Math.cos;
var mathACos = Math.acos;
var mathATan2 = Math.atan2;
// const mathAbs = Math.abs;
var mathSqrt = Math.sqrt;
var mathMax = Math.max;
var mathMin = Math.min;
var e = 1e-4;
// 注册 css 属性
var SECTOR_CSS_PROPERTY = [{
  name: 'sr',
  inherits: false,
  interpolable: true,
  syntax: gLite.PropertySyntax.LENGTH_PERCENTAGE
}, {
  name: 'sr0',
  inherits: false,
  interpolable: true,
  syntax: gLite.PropertySyntax.LENGTH_PERCENTAGE
}, {
  name: 'startAngle',
  inherits: false,
  interpolable: true,
  syntax: gLite.PropertySyntax.ANGLE
}, {
  name: 'endAngle',
  inherits: false,
  interpolable: true,
  syntax: gLite.PropertySyntax.ANGLE
}];
SECTOR_CSS_PROPERTY.forEach(function (property) {
  gLite.CSS.registerProperty(property);
});
var polarToCartesian = function polarToCartesian(centerX, centerY, radius, angleInRadian) {
  return {
    x: centerX + radius * Math.cos(angleInRadian),
    y: centerY + radius * Math.sin(angleInRadian)
  };
};
function intersect(x0, y0, x1, y1, x2, y2, x3, y3) {
  var dx10 = x1 - x0;
  var dy10 = y1 - y0;
  var dx32 = x3 - x2;
  var dy32 = y3 - y2;
  var t = dy32 * dx10 - dx32 * dy10;
  if (t * t < e) {
    return;
  }
  t = (dx32 * (y0 - y2) - dy32 * (x0 - x2)) / t;
  return [x0 + t * dx10, y0 + t * dy10];
}
// Compute perpendicular offset line of length rc.
function computeCornerTangents(x0, y0, x1, y1, radius, cr, clockwise) {
  var x01 = x0 - x1;
  var y01 = y0 - y1;
  var lo = (clockwise ? cr : -cr) / mathSqrt(x01 * x01 + y01 * y01);
  var ox = lo * y01;
  var oy = -lo * x01;
  var x11 = x0 + ox;
  var y11 = y0 + oy;
  var x10 = x1 + ox;
  var y10 = y1 + oy;
  var x00 = (x11 + x10) / 2;
  var y00 = (y11 + y10) / 2;
  var dx = x10 - x11;
  var dy = y10 - y11;
  var d2 = dx * dx + dy * dy;
  var r = radius - cr;
  var s = x11 * y10 - x10 * y11;
  var d = (dy < 0 ? -1 : 1) * mathSqrt(mathMax(0, r * r * d2 - s * s));
  var cx0 = (s * dy - dx * d) / d2;
  var cy0 = (-s * dx - dy * d) / d2;
  var cx1 = (s * dy + dx * d) / d2;
  var cy1 = (-s * dx + dy * d) / d2;
  var dx0 = cx0 - x00;
  var dy0 = cy0 - y00;
  var dx1 = cx1 - x00;
  var dy1 = cy1 - y00;
  // Pick the closer of the two intersection points
  // TODO: Is there a faster way to determine which intersection to use?
  if (dx0 * dx0 + dy0 * dy0 > dx1 * dx1 + dy1 * dy1) {
    cx0 = cx1;
    cy0 = cy1;
  }
  return {
    cx: cx0,
    cy: cy0,
    x0: -ox,
    y0: -oy,
    x1: cx0 * (radius / r - 1),
    y1: cy0 * (radius / r - 1)
  };
}
function computeArcSweep(startAngle, endAngle) {
  endAngle = endAngle < 0 && startAngle >= 0 ? endAngle + PI2 : endAngle;
  return endAngle - startAngle <= PI ? 0 : 1;
}
var Sector = /*#__PURE__*/function (_Path) {
  _inheritsLoose(Sector, _Path);
  // parsedStyle: any;
  function Sector(config) {
    var _this;
    _this = _Path.call(this, config) || this;
    _this.updatePath();
    return _this;
  }
  var _proto = Sector.prototype;
  _proto.setAttribute = function setAttribute(name, value, force) {
    _Path.prototype.setAttribute.call(this, name, value, force);
    if (['startAngle', 'endAngle', 'sr', 'sr0', 'radius'].indexOf(name) > -1) {
      this.updatePath();
    }
  };
  _proto.updatePath = function updatePath() {
    // @ts-ignore
    var _this$parsedStyle = this.parsedStyle,
      x = _this$parsedStyle.x,
      y = _this$parsedStyle.y,
      startAngle = _this$parsedStyle.startAngle,
      endAngle = _this$parsedStyle.endAngle,
      sr = _this$parsedStyle.sr,
      sr0 = _this$parsedStyle.sr0,
      radius = _this$parsedStyle.radius;
    var path = this.createPath(x, y, startAngle ? gLite.deg2rad(startAngle) : 0, endAngle ? gLite.deg2rad(endAngle) : Math.PI * 2, sr ? sr : 0, sr0 ? sr0 : 0, radius ? radius : [0, 0, 0, 0]);
    _Path.prototype.setAttribute.call(this, 'path', path);
  };
  _proto.createPath = function createPath(x, y, startAngle, endAngle, r, r0, borderRadius) {
    if (r <= 0) {
      return;
    }
    var start = polarToCartesian(x, y, r, startAngle);
    var end = polarToCartesian(x, y, r, endAngle);
    var innerStart = polarToCartesian(x, y, r0, startAngle);
    var innerEnd = polarToCartesian(x, y, r0, endAngle);
    // 整圆
    if (util.isNumberEqual(endAngle - startAngle, Math.PI * 2)) {
      // 整个圆是分割成两个圆
      var middlePoint = polarToCartesian(x, y, r, startAngle + Math.PI);
      var innerMiddlePoint = polarToCartesian(x, y, r0, startAngle + Math.PI);
      var circlePathCommands = [['M', start.x, start.y], ['A', r, r, 0, 1, 1, middlePoint.x, middlePoint.y], ['A', r, r, 0, 1, 1, end.x, end.y], ['M', innerStart.x, innerStart.y]];
      if (r0) {
        circlePathCommands.push(['A', r0, r0, 0, 1, 0, innerMiddlePoint.x, innerMiddlePoint.y]);
        circlePathCommands.push(['A', r0, r0, 0, 1, 0, innerEnd.x, innerEnd.y]);
      }
      circlePathCommands.push(['M', start.x, start.y]);
      circlePathCommands.push(['Z']);
      return circlePathCommands;
    }
    var angle = endAngle - startAngle;
    var xrs = r * mathCos(startAngle);
    var yrs = r * mathSin(startAngle);
    var xire = r0 * mathCos(endAngle);
    var yire = r0 * mathSin(endAngle);
    var xre = r * mathCos(endAngle);
    var yre = r * mathSin(endAngle);
    var xirs = r0 * mathCos(startAngle);
    var yirs = r0 * mathSin(startAngle);
    // 顺时针反向，同 radius
    var outStartRadius = borderRadius[0],
      outEndRadius = borderRadius[1],
      innerEndRadius = borderRadius[2],
      innerStartRadius = borderRadius[3];
    var halfRadius = (r - r0) / 2;
    var outStartBorderRadius = mathMin(halfRadius, outStartRadius);
    var outEndBorderRadius = mathMin(halfRadius, outEndRadius);
    var innerEndBorderRadius = mathMin(halfRadius, innerEndRadius);
    var innerStartBorderRadius = mathMin(halfRadius, innerStartRadius);
    var outBorderRadiusMax = mathMax(outStartBorderRadius, outEndBorderRadius);
    var innerBorderRadiusMax = mathMax(innerEndBorderRadius, innerStartBorderRadius);
    var limitedOutBorderRadiusMax = outBorderRadiusMax;
    var limitedInnerBorderRadiusMax = innerBorderRadiusMax;
    // draw corner radius
    if (outBorderRadiusMax > e || innerBorderRadiusMax > e) {
      // restrict the max value of corner radius
      if (angle < PI) {
        var it = intersect(xrs, yrs, xirs, yirs, xre, yre, xire, yire);
        if (it) {
          var x0 = xrs - it[0];
          var y0 = yrs - it[1];
          var x1 = xre - it[0];
          var y1 = yre - it[1];
          var a = 1 / mathSin(mathACos((x0 * x1 + y0 * y1) / (mathSqrt(x0 * x0 + y0 * y0) * mathSqrt(x1 * x1 + y1 * y1))) / 2);
          var b = mathSqrt(it[0] * it[0] + it[1] * it[1]);
          limitedOutBorderRadiusMax = mathMin(outBorderRadiusMax, (r - b) / (a + 1));
          limitedInnerBorderRadiusMax = mathMin(innerBorderRadiusMax, (r0 - b) / (a - 1));
        }
      }
    }
    var arcSweep = computeArcSweep(startAngle, endAngle);
    var clockwise = true;
    var sectorPathCommands = [];
    if (limitedOutBorderRadiusMax > e) {
      var crStart = mathMin(outStartRadius, limitedOutBorderRadiusMax);
      var crEnd = mathMin(outEndRadius, limitedOutBorderRadiusMax);
      var ct0 = computeCornerTangents(xirs, yirs, xrs, yrs, r, crStart, clockwise);
      var ct1 = computeCornerTangents(xre, yre, xire, yire, r, crEnd, clockwise);
      sectorPathCommands.push(['M', x + ct0.cx + ct0.x0, y + ct0.cy + ct0.y0]);
      // Have the corners merged?
      if (limitedOutBorderRadiusMax < outBorderRadiusMax && crStart === crEnd) {
        var outStartBorderRadiusStartAngle = mathATan2(ct0.y0, ct0.x0);
        var outStartBorderRadiusEndAngle = mathATan2(ct1.y0, ct1.x0);
        sectorPathCommands.push(['A', limitedOutBorderRadiusMax, limitedOutBorderRadiusMax, 0, computeArcSweep(outStartBorderRadiusStartAngle, outStartBorderRadiusEndAngle), 1, x + ct1.cx + ct1.x0, y + ct1.cy + ct1.y0]);
      } else {
        // draw the two corners and the ring
        if (crStart > 0) {
          var _outStartBorderRadiusStartAngle = mathATan2(ct0.y0, ct0.x0);
          var _outStartBorderRadiusEndAngle = mathATan2(ct0.y1, ct0.x1);
          var outStartBorderRadiusEndPoint = polarToCartesian(x, y, r, _outStartBorderRadiusEndAngle);
          sectorPathCommands.push(['A', crStart, crStart, 0, computeArcSweep(_outStartBorderRadiusStartAngle, _outStartBorderRadiusEndAngle), 1, outStartBorderRadiusEndPoint.x, outStartBorderRadiusEndPoint.y]);
        }
        var outRadiusStartAngle = mathATan2(ct0.cy + ct0.y1, ct0.cx + ct0.x1);
        var outRadiusEndAngle = mathATan2(ct1.cy + ct1.y1, ct1.cx + ct1.x1);
        var outRadiusEndPoint = polarToCartesian(x, y, r, outRadiusEndAngle);
        sectorPathCommands.push(['A', r, r, 0, computeArcSweep(outRadiusStartAngle, outRadiusEndAngle), 1, outRadiusEndPoint.x, outRadiusEndPoint.y]);
        if (crEnd > 0) {
          var outEndBorderRadiusStartAngle = mathATan2(ct1.y1, ct1.x1);
          var outEndBorderRadiusEndAngle = mathATan2(ct1.y0, ct1.x0);
          sectorPathCommands.push(['A', crEnd, crEnd, 0, computeArcSweep(outEndBorderRadiusStartAngle, outEndBorderRadiusEndAngle), 1, x + ct1.cx + ct1.x0, y + ct1.cy + ct1.y0]);
        }
      }
    } else {
      sectorPathCommands.push(['M', start.x, start.y]);
      sectorPathCommands.push(['A', r, r, 0, arcSweep, 1, end.x, end.y]);
    }
    // no inner ring, is a circular sector
    if (r0 < e) {
      sectorPathCommands.push(['L', innerEnd.x, innerEnd.y]);
    }
    // the inner ring has corners
    else if (limitedInnerBorderRadiusMax > e) {
      var _crStart = mathMin(innerStartRadius, limitedInnerBorderRadiusMax);
      var _crEnd = mathMin(innerEndRadius, limitedInnerBorderRadiusMax);
      var _ct = computeCornerTangents(xire, yire, 0, 0, r0 - r, _crEnd, clockwise);
      var _ct2 = computeCornerTangents(0, 0, xirs, yirs, r0 - r, _crStart, clockwise);
      sectorPathCommands.push(['L', x + _ct.cx + _ct.x0, y + _ct.cy + _ct.y0]);
      // Have the corners merged?
      if (limitedInnerBorderRadiusMax < innerBorderRadiusMax && _crStart === _crEnd) {
        var innerStartBorderRadiusStartAngle = mathATan2(_ct.y0, _ct.x0);
        var innerStartBorderRadiusEndAngle = mathATan2(_ct2.y0, _ct2.x0);
        var innerStartBorderRadiusEndPoint = polarToCartesian(x, y, r0, innerStartBorderRadiusEndAngle);
        sectorPathCommands.push(['A', limitedOutBorderRadiusMax, limitedOutBorderRadiusMax, 0, computeArcSweep(innerStartBorderRadiusStartAngle, innerStartBorderRadiusEndAngle), 1, innerStartBorderRadiusEndPoint.x, innerStartBorderRadiusEndPoint.y]);
      }
      // draw the two corners and the ring
      else {
        if (_crEnd > 0) {
          var _innerStartBorderRadiusStartAngle = mathATan2(_ct.y0, _ct.x0);
          var _innerStartBorderRadiusEndAngle = mathATan2(_ct.y1, _ct.x1);
          var _innerStartBorderRadiusEndPoint = polarToCartesian(x, y, r0 - r, _innerStartBorderRadiusEndAngle);
          sectorPathCommands.push(['A', _crEnd, _crEnd, 0, computeArcSweep(_innerStartBorderRadiusStartAngle, _innerStartBorderRadiusEndAngle), 1, _innerStartBorderRadiusEndPoint.x, _innerStartBorderRadiusEndPoint.y]);
        }
        var innerRadiusStartAngle = mathATan2(_ct.cy + _ct.y1, _ct.cx + _ct.x1);
        var innerRadiusEndAngle = mathATan2(_ct2.cy + _ct2.y1, _ct2.cx + _ct2.x1);
        var innerRadiusEndPoint = polarToCartesian(x, y, r0, innerRadiusEndAngle);
        sectorPathCommands.push(['A', r0, r0, 0, computeArcSweep(innerRadiusEndAngle, innerRadiusStartAngle), 0, innerRadiusEndPoint.x, innerRadiusEndPoint.y]);
        sectorPathCommands.push(['L', innerRadiusEndPoint.x, innerRadiusEndPoint.y]);
        if (_crStart > 0) {
          var innerEndBorderRadiusStartAngle = mathATan2(_ct2.y1, _ct2.x1);
          var innerEndBorderRadiusEndAngle = mathATan2(_ct2.y0, _ct2.x0);
          sectorPathCommands.push(['A', _crStart, _crStart, 0, computeArcSweep(innerEndBorderRadiusStartAngle, innerEndBorderRadiusEndAngle), 1, x + _ct2.cx + _ct2.x0, y + _ct2.cy + _ct2.y0]);
        }
      }
    }
    // the inner ring is just a circular arc
    else {
      sectorPathCommands.push(['L', innerEnd.x, innerEnd.y]);
      sectorPathCommands.push(['A', r0, r0, 0, arcSweep, 0, innerStart.x, innerStart.y]);
    }
    sectorPathCommands.push(['Z']);
    return sectorPathCommands;
  };
  return Sector;
}(gLite.Path);

exports.Arrow = Arrow;
exports.Sector = Sector;
